﻿/*Copyright 2024 Wojciech Krzysztof Fiałkiewicz

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Windows.Forms;

namespace Fractals_Visualizer
{
    public abstract partial class Fractal : Form, IWindowsCount, IFractal
    {
        public delegate void ClosingFractal();
        public event ClosingFractal ClosingWindow;

        private ComplexNumber m_C;
        private Bitmap m_Bmp;

        private bool m_DrawingZoom;
        private Point m_PosBeg;
        private Point m_Pos;
        private bool m_Resizing;
        private bool m_DrawRectangle;
        private Point m_MousePos;

        protected double MinVertical = -Math.PI;
        protected double MaxVertical = Math.PI;
        protected double MinHorizontal = -Math.PI;
        protected double MaxHorizontal = Math.PI;
        protected int Amount = 30;
        public Fractal()
        {
            InitializeComponent();
        }

        public void SetNumber(int nr)
        {
            this.Text = this.Text + ": " + nr.ToString();
        }

        private void Fractal_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (ClosingWindow != null)
            {
                ClosingWindow();
            }
        }

        public void DrawFractal(ComplexNumber c)
        {
            m_C = c;
            DrawFractal();
        }

        public void DrawFractal()
        {
            m_Bmp = GetScreenBitmap(panelFractal.Width, panelFractal.Height);
            panelFractal.Refresh();
        }

        public void SetClientSize(int width, int height)
        {
            base.ClientSize = new Size(width, height);
        }

        public void SetDeep(int amount)
        {
            this.Amount = amount;
        }

        public void Save()
        {
            SaveFileDialog sfd = new SaveFileDialog();

            sfd.Filter = "JPeg Image|*.jpg";
            sfd.Title = "Save an Image File";
            
            if(sfd.ShowDialog() == DialogResult.OK)
            {
                m_Bmp.Save(sfd.FileName);
            }           
        }

        public void SaveCustomSize(int width, int height)
        {
            SaveFileDialog sfd = new SaveFileDialog();

            sfd.Filter = "JPeg Image|*.jpg";
            sfd.Title = "Save an Image File";

            m_Bmp = GetScreenBitmap(width, height);

            if (sfd.ShowDialog() == DialogResult.OK)
            {
                m_Bmp.Save(sfd.FileName);
            }
        }

        public int GetDeep()
        {
            return Amount;
        }

        public int GetWidth()
        {
            return panelFractal.Width;
        }

        public int GetHeight()
        {
            return panelFractal.Height;
        }

        protected abstract DisplayNumber DetermineNumber(ComplexNumber z, ComplexNumber c, int x, int y);

        private Bitmap GetScreenBitmap(int width, int height)
        {//TODO: Add Background Worker and Progress Display
            Bitmap bmp = new Bitmap(width, height);
            List<DisplayNumber> list = new List<DisplayNumber>();

            for (int x = 0; x < width; x++)
            {
                for (int y = 0; y < height; y++)
                {
                    double a = ConvertScreenToComplexHorizontal(y, height);
                    double b = ConvertScreenToComplexVertical(x, width);

                    ComplexNumber z = new ComplexNumber();

                    z.a = a;
                    z.b = b;

                    list.Add(DetermineNumber(z, m_C, x, y));
                }
            }

            foreach (DisplayNumber dn in list)
            {
                double color = ((double)dn.nr)/((double)dn.Nr);

                if (double.IsNaN(color) || double.IsNegativeInfinity(color) || double.IsPositiveInfinity(color) || double.IsInfinity(color))
                {
                    color = 1.0;
                }
                if (color > 1.0)
                {
                    color = 1.0;
                }
                else if (color < 0.0)
                {
                    color = 0.0;
                }
                bmp.SetPixel(dn.x, dn.y, Color.FromArgb((int)(255 * color), (int)(255 * color), (int)(255 * color)));
            }
            return bmp;
        }

        private Image GetScreenImage(int width, int height)
        {
            if (m_Bmp != null)
            {
                MemoryStream ms = new MemoryStream();
                Image img;
                Bitmap bmp;

                m_Bmp.Save(ms, ImageFormat.Jpeg);
                ms.Seek(0, SeekOrigin.Begin);
                img = Image.FromStream(ms);
                bmp = new Bitmap(img, new Size(width, height));
                ms = new MemoryStream();
                bmp.Save(ms, ImageFormat.Jpeg);
                ms.Seek(0, SeekOrigin.Begin);
                img = Image.FromStream(ms);

                return img;
            }
            else
            {
                return null;
            }
        }

        private Point GetPoint()
        {
            Point pt = new Point();

            if (m_Pos.X - m_PosBeg.X >= m_Pos.Y - m_PosBeg.Y)
            {
                pt.X = m_Pos.X;
                pt.Y = m_PosBeg.Y + (m_Pos.X - m_PosBeg.X);
            }
            else
            {
                pt.X = m_PosBeg.X + (m_Pos.Y - m_PosBeg.Y);
                pt.Y = m_Pos.Y;
            }
            return pt;
        }

        private void panelFractal_Paint(object sender, PaintEventArgs e)
        {
            if (m_Bmp != null)
            {
                Panel p = (sender as Panel);
                Graphics g = p.CreateGraphics();
                Image img = GetScreenImage(panelFractal.Width, panelFractal.Height);

                g.DrawImage(img, new Point(0, 0));
                if (m_DrawRectangle)
                {
                    Point pt = new Point();

                    pt = GetPoint();
                    g.DrawRectangle(Pens.Black, m_PosBeg.X, m_PosBeg.Y, pt.X - m_PosBeg.X, pt.Y - m_PosBeg.Y);
                    m_DrawRectangle = false;
                }
                g.Dispose();
            }
        }

        private double ConvertScreenToComplexVertical(int x, int width)
        {
            return Tools.ConvertScreenToComplex(x, width, MinVertical, MaxVertical);
        }

        private double ConvertScreenToComplexHorizontal(int y, int height)
        {
            return Tools.ConvertScreenToComplex(y, height, MinHorizontal, MaxHorizontal);
        }

        private void Fractal_ResizeEnd(object sender, EventArgs e)
        {
            if (m_Resizing)
            {
                DrawFractal(m_C);
                m_Resizing = false;
            }
        }

        private void Fractal_ResizeBegin(object sender, EventArgs e)
        {
            m_Resizing = true;
        }

        private void panelFractal_Click(object sender, EventArgs e)
        {
            if (m_DrawingZoom)
            {
                m_Pos = m_MousePos;
                UpdateComplexArea(m_PosBeg, GetPoint());
                DrawFractal(m_C);
                m_DrawRectangle = false;
                m_DrawingZoom = false;
            }
            else
            {
                m_PosBeg = m_MousePos;
                m_DrawingZoom = true;
            }
        }

        private void panelFractal_MouseMove(object sender, MouseEventArgs e)
        {
            m_MousePos = e.Location;
            if (m_DrawingZoom)
            {
                m_Pos = m_MousePos;
                m_DrawRectangle = true;
                panelFractal.Refresh();
                m_DrawRectangle = false;
            }
        }

        private void UpdateComplexArea(Point beg, Point end)
        {
            double vmin, vmax, hmin, hmax;

            if (beg.X < end.X)
            {
                vmin = Tools.ConvertScreenToComplex(beg.X, panelFractal.Width, MinVertical, MaxVertical);
                vmax = Tools.ConvertScreenToComplex(end.X, panelFractal.Width, MinVertical, MaxVertical);
            }
            else
            {
                vmin = Tools.ConvertScreenToComplex(end.X, panelFractal.Width, MinVertical, MaxVertical);
                vmax = Tools.ConvertScreenToComplex(beg.X, panelFractal.Width, MinVertical, MaxVertical);
            }
            if (beg.Y < end.Y)
            {
                hmin = Tools.ConvertScreenToComplex(beg.Y, panelFractal.Height, MinHorizontal, MaxHorizontal);
                hmax = Tools.ConvertScreenToComplex(end.Y, panelFractal.Height, MinHorizontal, MaxHorizontal);
            }
            else
            {
                hmin = Tools.ConvertScreenToComplex(end.Y, panelFractal.Height, MinHorizontal, MaxHorizontal);
                hmax = Tools.ConvertScreenToComplex(beg.Y, panelFractal.Height, MinHorizontal, MaxHorizontal);
            }
            MinVertical = vmin;
            MaxVertical = vmax;
            MinHorizontal = hmin;
            MaxHorizontal = hmax;
        }
    }
}
