﻿/*Copyright 2024 Wojciech Krzysztof Fiałkiewicz

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/

using System;
using System.Drawing;
using System.Windows.Forms;

namespace Fractals_Visualizer
{
    public partial class FractalProperties : Form, IWindowsCount
    {
        private const double HorizontalMin = -Math.PI / 3.0;
        private const double HorizontalMax = Math.PI / 3.0;
        private const double VerticalMin = -Math.PI / 3.0;
        private const double VerticalMax = Math.PI / 3.0;

        private Fractal m_Fractal;
        private int m_Nr;
        protected double m_Re;
        protected double m_Im;
        private double m_ReGrey;
        private double m_ImGrey;
        private bool m_ShowGrey;
        private int m_MouseX;
        private int m_MouseY;

        public FractalProperties()
        {
            InitializeComponent();
            m_Re = 0.0;
            m_Im = 0.0;
            m_ReGrey = 0.0;
            m_ImGrey = 0.0;
            m_ShowGrey = false;
            textBoxRe.Text = m_Re.ToString();
            textBoxIm.Text = m_Im.ToString();
            this.DoubleBuffered = true;
        }

        public void SetNumber(int nr)
        {
            m_Nr = nr;
            this.Text = this.Text + ": " + m_Nr.ToString();
        }

        protected virtual Fractal CreateFractal()
        {
            return null;
        }

        private void FractalProperties_Activated(object sender, System.EventArgs e)
        {
            if (m_Fractal == null)
            {
                m_Fractal = CreateFractal();
                m_Fractal.MdiParent = this.MdiParent;
                (m_Fractal as IWindowsCount).SetNumber(m_Nr);
                m_Fractal.ClosingWindow += M_Fractal_ClosingWindow;
                m_Fractal.Show();
            }
        }

        private void M_Fractal_ClosingWindow()
        {
            m_Fractal = null;
        }

        private void panelComplexC_Paint(object sender, PaintEventArgs e)
        {
            Graphics g = (sender as Panel).CreateGraphics();

            g.FillRectangle(Brushes.White, (sender as Panel).DisplayRectangle);

            if (m_ShowGrey)
            {
                g.DrawLine(Pens.LightGray, CalculateHorizontal(HorizontalMin), CalculateHorizontal(m_ReGrey), CalculateHorizontal(HorizontalMax), CalculateHorizontal(m_ReGrey));
                g.DrawLine(Pens.LightGray, CalculateVertical(m_ImGrey), CalculateVertical(VerticalMin), CalculateVertical(m_ImGrey), CalculateVertical(VerticalMax));
            }

            g.DrawLine(Pens.Black, CalculateHorizontal(HorizontalMin), CalculateHorizontal(m_Re), CalculateHorizontal(HorizontalMax), CalculateHorizontal(m_Re));
            g.DrawLine(Pens.Black, CalculateVertical(m_Im), CalculateVertical(VerticalMin), CalculateVertical(m_Im), CalculateVertical(VerticalMax));

            g.Dispose();
        }

        private float CalculateVertical(double pos)
        {
            float ret;

            ret = Calculate(VerticalMin, VerticalMax, pos) * panelComplexC.Width;

            return ret;
        }

        private float CalculateHorizontal(double pos)
        {
            double p = Calculate(HorizontalMin, HorizontalMax, pos);
            double pMax = Calculate(HorizontalMin, HorizontalMax, HorizontalMax);
            float ret;

            ret = (float)((pMax - p) * panelComplexC.Width);

            return ret;
        }

        private float Calculate(double min, double max, double posInComplex)
        {
            float ret;

            ret = (float)((posInComplex - min) / (max - min));

            return ret;
        }

        private void panelComplexC_MouseEnter(object sender, EventArgs e)
        {
            m_ShowGrey = true;
        }

        private void panelComplexC_MouseLeave(object sender, EventArgs e)
        {
            m_ShowGrey = false;
            panelComplexC.Refresh();
        }

        private void panelComplexC_MouseMove(object sender, MouseEventArgs e)
        {
            m_MouseX = e.X;
            m_MouseY = e.Y;
            m_ReGrey = Tools.ConvertScreenToComplex(panelComplexC.Height - m_MouseY, panelComplexC.Height, HorizontalMin, HorizontalMax);
            m_ImGrey = Tools.ConvertScreenToComplex(m_MouseX, panelComplexC.Width, VerticalMin, VerticalMax);
            textBoxReGrey.Text = m_ReGrey.ToString();
            textBoxImGrey.Text = m_ImGrey.ToString();
            panelComplexC.Refresh();
        }

        private void panelComplexC_Click(object sender, EventArgs e)
        {
            m_Re = Tools.ConvertScreenToComplex(panelComplexC.Height - m_MouseY, panelComplexC.Height, HorizontalMin, HorizontalMax);
            m_Im = Tools.ConvertScreenToComplex(m_MouseX, panelComplexC.Width, VerticalMin, VerticalMax);
            textBoxRe.Text = m_Re.ToString();
            textBoxIm.Text = m_Im.ToString();
            panelComplexC.Refresh();

            DrawFractal();
        }

        private void textBoxRe_Validating(object sender, System.ComponentModel.CancelEventArgs e)
        {
            double val;

            if (double.TryParse(textBoxRe.Text, out val))
            {
                m_Re = val;
                e.Cancel = false;
                panelComplexC.Refresh();
                DrawFractal();
            }
            else
            {
                e.Cancel = true;
            }
        }

        private void textBoxIm_Validating(object sender, System.ComponentModel.CancelEventArgs e)
        {
            double val;

            if (double.TryParse(textBoxIm.Text, out val))
            {
                m_Im = val;
                e.Cancel = false;
                panelComplexC.Refresh();
                DrawFractal();
            }
            else
            {
                e.Cancel = true;
            }
        }

        private void DrawFractal()
        {
            ComplexNumber c = new ComplexNumber();

            c.a = m_Re;
            c.b = m_Im;

            m_Fractal.DrawFractal(c);
        }
    }
}
